from ._misc import docstring, MultiType
from ._api import AbstractAPI
import parse
from bottle import FormsDict


class API(AbstractAPI):
	def init(self,parsedoc=docstring):
		self.parsedoc = parsedoc

		self.classes = {}
		self.objects = {}
		self.methods = {} #not referring to object methods, but to http methods
		self.functions = {} #now these are the object methods .__.
		self.all_functions = [] #unbound


	# returns just the import information of this API
	def api_info(self):
		return {
			"url":self.pathprefix,
			"type":"restapi",
			"classes":[
				{
					"name":cls,
					"allowed":self.methods[self.classes[cls]],
					"instances":[name for name in self.objects[self.classes[cls]]],
					"methods":[],
					"attributes":{}
				} for cls in self.classes
			]
		}




	def handle(self,nodes,reqmethod,querykeys,headers):


		cls = self.classes[nodes.pop(0)]

		# REST access to resource
		if (reqmethod in ["POST"] and len(nodes) == 0) or (len(nodes) == 1):

			if reqmethod in ["POST"]:
				newobj = cls(**querykeys)
				# new object will sign up by itself thanks to class decorator
				return newobj.__apidict__()

			if reqmethod in ["GET","PATCH","DELETE"]:
				objkey = nodes.pop(0)
				obj = self.objects[cls][objkey]

			if reqmethod == "PATCH":
				try:
					return obj.__patch__(**querykeys)
				except:
					pass

				# if no __patch__ method is provided, we try everythin we can
				# to somehow patch this object
				for key in querykeys:
					for assignattempt in (
						# oh boy here we go again
						"getattr(obj,'set_' + key)(querykeys[key])",
						"getattr(obj,'set' + key)(querykeys[key])",
						"obj[key] = querykeys[key]",
						"setattr(obj,key,querykeys[key])"
					):
						try:
							exec(assignattempt)
							break
						except:
							pass

				return obj.__apidict__()

			if reqmethod == "GET":
				return obj.__apidict__()

			if reqmethod == "DELETE":
				del self.objects[cls][objkey]
				return {"status":"success"}

		# access to object methods
		elif len(nodes) > 1:
			objkey = nodes.pop(0)
			obj = self.objects[cls][objkey]

			for f in self.functions[cls]:

				if f["method"] != reqmethod: continue

				pathkeys = FormsDict()

				# match against paths
				r = parse.parse(f["path"],"/".join(nodes))
				if r is not None:
					func = f["func"]
					for k in r.named:
						# set vars according to path match
						pathkeys[k] = r[k]

					types = func.__annotations__
					for k in pathkeys:
						if k in types:
							if isinstance(types[k],MultiType):
								subtype = types[k].elementtype
								pk = pathkeys[k].split("/")
								pathkeys[k] = [subtype(e) for e in pk]
							else:
								pathkeys[k] = types[k](pathkeys[k])

					for k in querykeys:
						if k in types:
							if isinstance(types[k],MultiType):
								subtype = types[k].elementtype
								qk = querykeys.getall(k)
								querykeys[k] = [subtype(e) for e in qk]
							else:
								querykeys[k] = types[k](querykeys[k])

					if f["headers"]:
						return func(obj,**querykeys,**pathkeys,**headers)
					else:
						return func(obj,**querykeys,**pathkeys)


		return {"status":"failure"}


	# decorators for object methods
	def get(self,path,pass_headers=False):
		def decorator(func):
			# save reference to this function
			self.all_functions.append(
				{
					"path":path,
					"method":"GET",
					"func":func,
					"headers":pass_headers
				}
			)
			# return it unchanged
			return func
		return decorator

	def post(self,path,pass_headers=False):
		def decorator(func):
			# save reference to this function
			self.all_functions.append(
				{
					"path":path,
					"method":"POST",
					"func":func,
					"headers":pass_headers
				}
			)
			# return it unchanged
			return func
		return decorator


	# decorator for the class
	def apiclass(self,path,get=True,post=True,patch=True,delete=True):

		def decorator(cls):

			# save reference to this class
			self.classes[path] = cls
			self.objects[cls] = {}
			self.methods[cls] = {"GET":get,"POST":post,"PATCH":patch,"DELETE":delete}
			self.functions[cls] = []

			original_init = cls.__init__

			def new_init(self2,*args,**kwargs):
				# init normally
				original_init(self2,*args,**kwargs)
				# then sign up
				self.objects[cls][self2.__apiname__] = self2
				# self is the api object, self2 the object being initialized here

			cls.__init__ = new_init

			# unbound functions do not know their class ahead of time,
			# so we save them temporarily and then assign them on class
			# registration

			attrs = [cls.__dict__[k] for k in cls.__dict__]
			# include superclasses
			for superclass in cls.__bases__:
				attrs += [superclass.__dict__[k] for k in superclass.__dict__]
			# check if any decorated functions are methods of this class
			for f in self.all_functions:
				if f["func"] in attrs:
					self.functions[cls].append(f)
					#del self.all_functions[name]

			# return
			return cls

		return decorator



import re

regex_param = re.compile(r":param (.*)? (.*)?: (.*)")
regex_return = re.compile(r":return: (.*)")
regex_rtype = re.compile(r":rtype: (.*)")

# default function to extract information from functions
def docstring(func):
	try:
		docstr = func.__doc__
		docstr = docstr.replace("\t","")

		params = {}
		returns = {}

		lines = docstr.split("\n")
		desc = []
		for l in lines:
			if l.startswith(":"):
				match = regex_param.match(l)
				if match is not None:
					type, name, d = match.groups()
					params[name] = {"type":type,"desc":d}
					continue
				match = regex_return.match(l)
				if match is not None:
					returns["desc"], = match.groups()
					continue
				match = regex_rtype.match(l)
				if match is not None:
					returns["type"], = match.groups()
					continue

			else:
				desc.append(l)

		desc
		desc = "\n".join(desc).strip("\n")

		return {"desc":desc,"params":params,"returns":returns}
	except:
		return {"desc":"","params":{},"returns":{}}




# type hint for multiple values
class MultiType:

	# so we can define something like Multi[int]
	def __getitem__(self,key):
		instance = self.__class__()
		instance.elementtype = key
		return instance

Multi = MultiType()
Multi.elementtype = str



from ._misc import docstring, MultiType
from ._api import AbstractAPI
import parse
from bottle import FormsDict


class API(AbstractAPI):
	def init(self,parsedoc=docstring):
		self.parsedoc = parsedoc

		#self.functions = FunctionHolder()
		self.functions = []
		# dict: route, method, func

	# returns just the import information of this API
	def api_info(self):
		return {
			"url":self.pathprefix,
			"type":"functionapi",
			"endpoints":[
				{
					"name":f["path"],
					"method":f["method"],
					"description":self.parsedoc(f["func"])["desc"],
					"parameters":self.parsedoc(f["func"])["params"],
					#"parameters":{
					#	param:{
					#		"type":str(self.functions[pth][0].__annotations__.get(param)),
					#		"desc":"tbd"
					#	}
					#for param in self.functions[pth][0].__code__.co_varnames},
					"returns":self.parsedoc(f["func"])["returns"]
				} for f in self.functions
			]
		}




	def handle(self,nodes,reqmethod,querykeys,headers):


		for f in self.functions:

			if f["method"] != reqmethod: continue

			pathkeys = FormsDict()

			# match against paths
			r = parse.parse(f["path"],"/".join(nodes))
			if r is not None:
				func = f["func"]
				for k in r.named:
					# set vars according to path match
					pathkeys[k] = r[k]

				types = func.__annotations__
				for k in pathkeys:
					if k in types:
						if isinstance(types[k],MultiType):
							subtype = types[k].elementtype
							pk = pathkeys[k].split("/")
							pathkeys[k] = [subtype(e) for e in pk]
						else:
							pathkeys[k] = types[k](pathkeys[k])

				for k in querykeys:
					if k in types:
						if isinstance(types[k],MultiType):
							subtype = types[k].elementtype
							qk = querykeys.getall(k)
							querykeys[k] = [subtype(e) for e in qk]
						else:
							querykeys[k] = types[k](querykeys[k])

				if f["headers"]:
					return func(**querykeys,**pathkeys,**headers)
				else:
					return func(**querykeys,**pathkeys)

		return {"error":"Not found"}


	def get(self,path,pass_headers=False):

		def decorator(func):
			self.functions.append(
				{
					"path":path,
					"method":"GET",
					"func":func,
					"headers":pass_headers
				}
			)

			# return function unchanged
			return func
		return decorator

	def post(self,path,pass_headers=False):

		def decorator(func):
			self.functions.append(
				{
					"path":path,
					"method":"POST",
					"func":func,
					"headers":pass_headers
				}
			)

			# return function unchanged
			return func
		return decorator



# Credit to https://gist.github.com/reimund/5435343/


def format(d, root_node="root"):
	return '<?xml version="1.0" encoding="UTF-8"?>' + formatelement(d,root_node)

def formatelement(d, root_node="root"):
	wrap          =     False if None == root_node or isinstance(d, list) else True
	root          = 'objects' if None == root_node else root_node
	root_singular = root[:-1] if 's' == root[-1] and None == root_node else root
	xml           = ''
	children      = []

	if isinstance(d, dict):
		for key, value in dict.items(d):
			if isinstance(value, dict):
				children.append(formatelement(value, key))
			elif isinstance(value, list):
				children.append(formatelement(value, key))
			else:
				xml = xml + ' ' + key + '="' + str(value) + '"'
	else:
		for value in d:
			children.append(formatelement(value, root_singular))

	end_tag = '>' if 0 < len(children) else '/>'

	if wrap or isinstance(d, dict):
		xml = '<' + root + xml + end_tag

	if 0 < len(children):
		for child in children:
			xml = xml + child

		if wrap or isinstance(d, dict):
			xml = xml + '</' + root + '>'

	return xml



name = "nimrodel"
version = 0,6,2
versionstr = ".".join(str(n) for n in version)
author = {
	"name": "Johannes Krattenmacher",
	"email": "python@krateng.dev",
	"github": "krateng"
}
requires = [
	"bottle",
	"waitress",
	"doreah>=1.2.1",
	"parse"
]




from ._apiobject import API as ObjectAPI
from ._apifunction import API as EndpointAPI
from ._apirest import API as RestAPI
from ._misc import Multi

# short names
OAPI = ObjectAPI
EAPI = EndpointAPI
RAPI = RestAPI

# default
API = OAPI



from ._misc import docstring
from ._api import AbstractAPI



class API(AbstractAPI):
	def init(self,parsedoc=docstring):

		self.parsedoc = parsedoc

		self.classes = {}
		self.objects = {}
		self.functions = {} #tuple function, method

		self.all_functions = []


	def api_info(self):
		return {
			"url":self.pathprefix,
			"type":"objectapi",
			"classes":[
				{
					"name":cls,
					"instances":[name for name in self.objects[self.classes[cls]]],
					"methods":[
						{
							"name":name,
							"method":self.functions[self.classes[cls]][name][1],
							"description":self.parsedoc(self.functions[self.classes[cls]][name][0])["desc"],
							"parameters":self.parsedoc(self.functions[self.classes[cls]][name][0])["params"],
							"returns":self.parsedoc(self.functions[self.classes[cls]][name][0])["returns"]
						} for name in self.functions[self.classes[cls]]
					]
				} for cls in self.classes
			]
		}

		# access methods
	#	dec = self.server.get(self.pathprefix + "/<classname>/<objectname>/<functionname>")
	#	self.route_to_function = dec(self.route_to_function)

		# access object itself
	#	dec = self.server.get(self.pathprefix + "/<classname>/<objectname>")
	#	self.route_to_object = dec(self.route_to_object)



	def handle(self,nodes,reqmethod,keys,headers):

		post_allowed = (reqmethod == "POST")

		cls = self.classes[nodes.pop(0)]
		obj = self.objects[cls][nodes.pop(0)]

		current = obj


		while len(nodes) > 0:
			next = nodes.pop(0)
			func,httpmethod = self.functions[current.__class__][next]
			if httpmethod == "GET":
				current = func(current,**keys)
			elif httpmethod == "POST" and post_allowed:
				current = func(current,**keys)
				post_allowed = False
			else:
				return {"error":"HTTP method"}

		if post_allowed:
			# if we made a post request and never used any post step
			return {"error":"HTTP method"}

		# all is done, return last object
		if callable(getattr(current,"__apidict__",None)):
			return current.__apidict__()
		else:
			return current



#	def route_to_function(self,classname,objectname,functionname):
#		keys = FormsDict.decode(request.query)
#		cls = self.classes[classname]
#		obj = self.objects[cls][objectname]
#		func = self.functions[cls][functionname]
#		return func(obj,**keys)
#
#	def route_to_object(self,classname,objectname):
#		keys = FormsDict.decode(request.query)
#		cls = self.classes[classname]
#		obj = self.objects[cls][objectname]
#		return obj.__apidict__(**keys)

	# decorator for the method
	def get(self,path):

		def decorator(func):
			# save reference to this function
			self.all_functions.append((path,func,"GET"))
			# return it unchanged
			return func

		return decorator

	def post(self,path):

		def decorator(func):
			# save reference to this function
			self.all_functions.append((path,func,"POST"))
			# return it unchanged
			return func

		return decorator

	# decorator for the class
	def apiclass(self,path):

		def decorator(cls):

			# save reference to this class
			self.classes[path] = cls
			self.objects[cls] = {}
			self.functions[cls] = {}

			original_init = cls.__init__

			def new_init(self2,*args,**kwargs):
				# init normally
				original_init(self2,*args,**kwargs)
				# then sign up
				self.objects[cls][self2.__apiname__] = self2
				# self is the api object, self2 the object being initialized here

			cls.__init__ = new_init

			# assign functions
			#self.functions[cls] = self.all_functions
			#self.all_functions = {}

			# unbound functions do not know their class ahead of time,
			# so we save them temporarily and then assign them on class
			# registration

			attrs = [cls.__dict__[k] for k in cls.__dict__]
			# include superclasses
			for superclass in cls.__bases__:
				attrs += [superclass.__dict__[k] for k in superclass.__dict__]
			# check if any decorated functions are methods of this class
			for (pth,func,method) in self.all_functions:
				if func in attrs:
					self.functions[cls][pth] = func,method
					#del self.all_functions[name]



			# return
			return cls

		return decorator

	# manually register an object
##	def register_object(self,obj,name):
#		self.objects[name] = obj



from bottle import Bottle, FormsDict, request, response, Response
from bottle import run as bottlerun
import waitress
from threading import Thread
from doreah.pyhp import file
from doreah.logging import log
import pkg_resources
from . import versionstr

__logmodulename__ = "nimrodel"

class AbstractAPI:

	def __init__(self,
			server=None,
			port=1337,IPv6=True,
			path="api",
			delay=False,
			auth=None,
			type="json",root_node=None,
			debug=False,
			**kwargs):

		self.path = path
		self.pathprefix = "" if path is None else ("/" + path)

		if auth is not None:
			self.auth = auth

		self.type = type.lower()
		self.rootnode = root_node

		self.debug = debug

		self.init(**kwargs)

		if delay:
			pass
		else:
			self.initserver(port=port,IPv6=IPv6,server=server)
			self.setup_explorer()
			self.setup_routing()

	def auth(self,request):
		return True

	def initserver(self,server,port,IPv6):

		if server is None:
			host = "::" if IPv6 else "0.0.0.0"
			port = port
			self.server = Bottle()
			self.server._apis = [self,]
			t = Thread(target=bottlerun,args=(self.server,),kwargs={"host":host,"port":port,"server":"waitress"})
			t.daemon = True
			t.start()
		else:
			try:
				server._apis.append(self)
			except:
				server._apis = [self]
			self.server = server

	# if delay has been specified on creation, we can now mount this api to a server
	def mount(self,server=None,port=1337,IPv6=True):

		self.initserver(port=port,IPv6=IPv6,server=server)
		self.setup_explorer()
		self.setup_routing()

	def setup_explorer(self):
		# API explorer
		exploredec = self.server.get("/api_info")
		exploredec(self.explorer)

		g_exploredec = self.server.get("/api_explorer")
		g_exploredec(self.gexplorer)

	def setup_routing(self):

		dec = self.server.get(self.pathprefix + "/<fullpath:path>")
		dec(self.route)
		dec = self.server.post(self.pathprefix + "/<fullpath:path>")
		dec(self.route)
		dec = self.server.route(self.pathprefix + "/<fullpath:path>",method="PATCH")
		dec(self.route)
		dec = self.server.put(self.pathprefix + "/<fullpath:path>")
		dec(self.route)
		dec = self.server.delete(self.pathprefix + "/<fullpath:path>")
		dec(self.route)

	def explorer(self):
		return {
			"information":{
				"nimrodel-version":versionstr
			},
			"apis":[
				api.api_info() for api in self.server._apis
			]
		}

	def gexplorer(self):
		folder = pkg_resources.resource_filename(__name__,"res/")
		#pyhpstr = pkg_resources.resource_string(__name__,"res/apiexplorer.pyhp")
		return file(folder + "/apiexplorer.pyhp",self.explorer())



	def route(self,fullpath):
		# preprocess all requests
		headers = request.headers

		keys = FormsDict.decode(request.query)

		if self.debug:
			log("Request to " + fullpath)
			for k in keys:
				log("\t" + k + " = " + keys.get(k))

		if request.get_header("Content-Type") is not None and "application/json" in request.get_header("Content-Type"):
			json = request.json if request.json is not None else {}
			keys.update(json)

		else:
			keys.update(FormsDict.decode(request.forms))

		#print(keys)

		nodes = fullpath.split("/")
		reqmethod = request.method

		if self.auth(request):
			result = self.handle(nodes,reqmethod,keys,headers)
			if isinstance(result,Response):
				return result
			else:
				result = serialize(result)
				result = format_output[self.type](result,root_node=self.rootnode)
				return result
		else:
			response.status = 403
			return "Access denied"


from .output import json, xml

format_output = {
	"json":json.format,
	"xml":xml.format
}


def serialize(obj):

	if isinstance(obj,str) or isinstance(obj,int): return obj
	if obj == [] or obj == {}: return obj

	try:
		return {k:serialize(obj[k]) for k in obj}
	except Exception as e:
		pass

	try:
		return [serialize(element) for element in obj]
	except Exception as e:
		pass

	for f in ["__apidict__","__json__","__dict__"]:
		try:
			return serialize(getattr(obj,f)())
		except Exception as e:
			pass

	return obj
